<?php
/**
 * AnswerDAO handles the CRUD actions of student answers.
 */
class AnswerDAO extends DAO
{
    /**
     * table: Holds the DB table name, including the prefix
     * @var string
     */
    private static $table;
    /**
     * processErrors: For error handling and escalation
     * @var string
     */
    private $processErrors = "";
    /**
     * Summary of processSummary: Concatenation of $processErrors
     * @var string
     */
    private $processSummary = "";
    /**
     * Summary of __construct
     */
    public function __construct()
    {
        $table = CFG::mkTableName("answers");
        self::$table = $table;
    }
    /**
     * A refactorable, general-purpose DAO method (R in CRUD)
     * @param [] $where Key-value pairs that will become an SQL WHERE clause
     * @param bool $verbose Verbosity of error messages
     * @return Answer[]
     */
    public function get($where = [], $verbose = false)
    {
        $table = self::$table;
        $answers = [];
        $rows = DB::$db->get(
            table: $table,
            where: $where
        );
        $got = count($rows);
        if ($verbose && $got <= 0) { // Not found in the DB
            HT::error("<strong>AnswerDAO</strong>: 
                No answers in the DB!<br>
                Run AnswerDAO->process(student) for each student first.");
        } else {
            foreach ($rows as $row) {
                $answers[] = new Answer(
                    $row['student_email'],
                    $row['subquestion_name'],
                    $row['filename'],
                    $row['fulltext'],
                    $row['hash0'],
                    $row['hash1'],
                    $row['comment']
                );
            }
        }
        return $answers;
    }
    /**
     * Get Answers by specified object(s)
     * @param Student $student
     * @param SubQuestion $subQuestion
     * @param Answer $answer
     * @return Answer[]
     */
    public function getByObject($student, $subQuestion = null, $answer = null)
    {
        $where = ['student_email' => $student->getEmail()];
        if ($subQuestion) {
            $where['subquestion_name'] = $subQuestion->getName();
        }
        $got = $this->get($where);
        if ($answer) { // Use the short file name to filter, not 
            $answers = [];
            foreach ($got as $a) {
                if ($answer->getShortFileName() == $a->getShortFileName()) {
                    $answers[] = $a;
                }
            }
        } else {
            $answers = $got;
        }
        // Sort the answers by short file name, for aesthetics
        $answers = self::sortByShortFileName($answers);
        return $answers;
    }
    /**
     * Return Answer objects based on $fileName
     * @param string $fileName
     * @return Answer[]
     */
    public function getByFileName($fileName)
    {
        $where = ['filename' => $fileName];
        return $this->get($where);
    }
    /**
     * Get the zip file that the student has uploaded.
     * (Or the one that was created by Grader, if finals.)
     * Also sets the zipFile property of $student.
     * @param Student $student
     * @return string
     */
    private function getStudentZip($student)
    {
        $zipFile = "";
        // Try the emailed zips first
        $email = $student->getEmail();
        $section = $student->getSection();
        $name = $student->getName();
        $leadIn = $student->mkLeadIn();
        $gPat = CFG::$emailFolder . "*$email*.zip";
        $fNames = glob($gPat);
        $nFiles = count($fNames);
        if ($nFiles == 0) { // Try the section folder
            $gPat = CFG::getSubmissionFolder($section) . "*$email*.zip";
            $fNames = glob($gPat);
            $nFiles = count($fNames);
        }
        if ($nFiles == 0) { // Try the section folder with student name
            $gPat = CFG::getSubmissionFolder($section) . "*$name*.zip";
            $fNames = glob($gPat);
            $nFiles = count($fNames);
        }
        if ($nFiles > 1) {
            $this->processErrors .= HT::warn(
                "$leadIn Found $nFiles Zip files. Will use the latest one.",
                true
            );
            // Ref: https://stackoverflow.com/a/35925596
            usort($fNames, function ($a, $b) {
                return filemtime($b) - filemtime($a);
            });
            $nFiles = 1;
        }
        if ($nFiles != 1) {
            $this->processErrors .= HT::error(
                "$leadIn Found $nFiles Zip files. Please investigate!",
                true
            );
            $student->setStatus("Absent")->update();
        } else {
            $zipFile = $fNames[0];
            $student->setZipFile($zipFile);
        }
        return $zipFile;
    }
    /**
     * Unzips the student upload 
     * @param Student $student
     * @return bool if succeeds
     */
    private function unzipStudent($student)
    {
        $zipFile = $student->getZipFile();
        $zip = new ZipArchive;
        $res = $zip->open($zipFile);
        if ($res === TRUE) {
            CFG::rmdir("foo");
            $zip->extractTo("foo");
        } else {
            $leadIn = $student->mkLeadIn();
            $this->processErrors .= HT::error(
                "$leadIn Error unzipping to <var>foo</var>!<br>
            &emsp;[<var>$zipFile</var>]",
                true
            );
            return false;
        }
        $zip->close();
        return true;
    }
    /**
     * Summary of cpAutoGraders
     * @param Student $student
     * @return bool
     */
    private function cpAutoGraders($student)
    { // No error handling for now!
        $answerFolder = $student->getAnswerFolder();
        foreach (CFG::$autoGrade as $files => $graders) {
            foreach ($graders as $grader) {
                $from = CFG::$solutionFolder . $grader;
                $to = $answerFolder . "/" . $grader;
                CFG::cp($from, $to, $mkDir = true);
            }
        }
        return true;
    }
    /**
     * Summary of stripCommon
     * @param mixed $strings
     * @return array|string
     */
    private static function stripCommon($strings)
    {
        if (CFG::isQuiz()) {
            $stripped = str_replace("foo/", "", $strings);
        } else {
            // need to remove silly paths students have used
            $difPosition = 10000; // just a large number
            $count = count($strings);
            for ($i = 1; $i < $count; $i++) {
                $position = strspn($strings[$i - 1] ^ $strings[$i], "\0");
                $difPosition = min([$difPosition, $position]);
            }
            if ($strings[0][$difPosition - 1] == 'q') {
                $difPosition--;
            }
            $stripped = [];
            foreach ($strings as $string) {
                $stripped[] = substr($string, $difPosition);
            }
        }
        return $stripped;
    }
    /**
     * Summary of mvStudentAnswers
     * @param mixed $student
     * @return string[]
     */
    private function mvStudentAnswers($student)
    {
        // Iterate over the files in unzipped (in foo)
        $refFileDAO = new RefFileDAO();
        $answers = glob(
            "foo/{*,*/*,*/*/*,*/*/*/*,,*/*/*/*/*}/*.{php,html}",
            GLOB_BRACE
        );
        if (empty($answers)) {
            return [];
        }
        $shortNames = self::stripCommon($answers);
        $answerFolder = $student->getAnswerFolder($student);
        if (realpath($answerFolder)) {
            $student->delete();
        }
        @mkdir($answerFolder, 0777, true);
        $moved = [];
        foreach ($answers as $key => $a) {
            $shortFileName = $shortNames[$key];
            $refFile = $refFileDAO->getRefFile($shortFileName, 'resources');
            if (empty($refFile)) {
                $leadIn = $student->mkLeadIn();
                $this->processErrors .= HT::error(
                    "$leadIn Created an unknown file <code>$shortFileName</code>",
                    true
                );
                continue;
            }
            $refShortFileName = $refFile->getShortFileName();
            $target = $answerFolder . $refShortFileName;
            CFG::mv($a, $target);
            $moved[$refShortFileName] = $target;
        }
        CFG::rmdir("foo");
        return $moved;
    }
    // This feels like bad OO design: Need the subQuestion name in the answers 
    // table (to facilitate grading). But the only way to get it is to use 
    // the filename as a key in the ref_files table.
    /**
     * Summary of getSubQuestionName
     * @param mixed $shortFileName
     * @return mixed
     */
    private static function getSubQuestionName($shortFileName)
    {
        $rubricDao = new RubricDAO();
        $rubrics = $rubricDao->get([
            'short_filename' => $shortFileName
        ]);
        $subQuestionName = "";
        if (count($rubrics) > 0) {
            $subQuestionName = $rubrics[0]->getSubQuestionName();
        }
        return $subQuestionName;
    }

    /**
     * Save the answers of a given student in the DB
     * @param mixed $student
     * @return void
     */
    private function put($student, $answers)
    {
        $leadIn = $student->mkLeadIn();
        if (count($answers) <= 0) { // No answers found. An error!
            $this->processErrors .= HT::error(
                "$leadIn No answers found. 
            Probably absent or alias-zipping.",
                true
            );
        } else {
            $table = self::$table;
            $columns = [
                'student_email',
                'subquestion_name',
                'filename',
                'fulltext',
                'hash0',
                'hash1',
            ];
            $rows = [];
            foreach ($answers as $answer) {
                $studentEmail = $student->getEmail();
                $subQuestionName = $answer->getSubQuestionName();
                $fileName = $answer->getFileName();
                $fullText = DB::$db->getFileContent($fileName);
                $hash0 = Report::hash($fileName);
                $hash1 = Report::hash($fileName, false);
                $rows[] = [
                    $studentEmail,
                    $subQuestionName,
                    $fileName,
                    $fullText,
                    $hash0,
                    $hash1,
                ];
            }
            DB::$db->update($table, $columns, $rows);
        }
    }
    /**
     * Summary of process
     * @param mixed $student
     * @return void
     */
    public function process($student)
    {
        $name = $student->getName();
        $studentEmail = $student->getEmail();
        $section = $student->getSection();
        $msg = $student->mkLeadIn();
        if ($zipFile = $this->getStudentZip($student)) {
            $dirName = realpath(dirname($zipFile));
            $baseName = basename($zipFile);
            $msg .= HT::success("Located the zip file: <br>
            &ensp;Folder: <code>$dirName</code><br>
            &ensp;File: <code>$baseName</code>", true);
            if ($this->unzipStudent($student)) {
                $answerFiles = $this->mvStudentAnswers($student);
                $count = count($answerFiles);
                if ($count) {
                    $msg .= HT::success("$count files successfully unzipped.", true);
                } else {
                    $msg .= HT::error("No files unzipped.", true);
                }
                $answers = [];
                foreach ($answerFiles as $shortFileName => $fileName) {
                    $subQuestionName = self::getSubQuestionName($shortFileName);
                    if ($subQuestionName) {
                        $answers[] = new Answer(
                            $studentEmail,
                            $subQuestionName,
                            $fileName
                        );
                    }
                }
                $count = count($answers);
                if ($count) {
                    $msg .= HT::success("$count Answer objects successfully created.", true);
                } else {
                    $msg .= HT::error("No Answer objects created.", true);
                }
                $this->put($student, $answers);
                $count = DB::$db->affectedRows();
                if ($count) {
                    $msg .= HT::success("$count records went into the database.", true);
                } else {
                    $msg .= HT::error("No database rows affected.", true);
                }
                $count = $this->cpAutoGraders($student);
                if ($count) {
                    $msg .= HT::success("Attempted autograded question. Copied.", true);
                } else {
                    $msg .= HT::error("No autograders copied.", true);
                }
            }
        }
        $msg .= $this->processErrors;
        $this->processSummary .= $this->processErrors;
        $this->processErrors = "";
        HT::info($msg);
    }
    /**
     * Summary of update
     * @param mixed $answer
     * @return void
     */
    public function update($answer)
    {
        $table = self::$table;
        $columns = [
            'student_email',
            'subquestion_name',
            'filename',
            'fulltext',
            'hash0',
            'hash1',
            'comment'
        ];
        $studentEmail = $answer->getStudentEmail();
        $subQuestionName = $answer->getSubQuestionName();
        $fileName = $answer->getFileName();
        $fulltext = DB::$db->sanitize($answer->getFullText());
        $hash0 = $answer->getHash0();
        $hash1 = $answer->getHash1();
        $comment = $answer->getComment();
        $rows = [
            $studentEmail,
            $subQuestionName,
            $fileName,
            $fulltext,
            $hash0,
            $hash1,
            $comment
        ];
        // DB::$db->put($table, $columns, $rows);
        DB::$db->update($table, $columns, $rows);
    }

    /**
     * @return mixed
     */
    public function getProcessSummary()
    {
        return $this->processSummary;
    }

    /**
     * @param mixed $processSummary 
     * @return self
     */
    public function setProcessSummary($processSummary): self
    {
        $this->processSummary = $processSummary;
        return $this;
    }
}