<?php
class RefFileDAO extends DAO
{
    private static $table;
    private static $gradables = [];
    // list of files that are different between solutions and resources (=> Gradable)
    public function __construct()
    {
        $table = CFG::mkTableName("ref_files");
        self::$table = $table;
    }
    public function get($where = [], $verbose = false)
    {
        $table = self::$table;
        $rows = DB::$db->get(
            table: $table,
            where: $where
        );
        $got = count($rows);
        if ($verbose && $got <= 0) { // Not found in the DB
            HT::error("<strong>RefFileDAO</strong>: 
                No resources or solutions in the DB!<br>
                Run RefFileDAO->process(student) for each student first.");
            return;
        } else {
            $refFiles = [];
            foreach ($rows as $row) {
                $refFile = new RefFile(
                    $row['subquestion_name'],
                    $row['filetype'],
                    $row['filename'],
                    $row['fulltext'],
                    $row['hash0'],
                    $row['hash1']
                );
                $refFile->setSim($row['sim']);
                $refFile->setGradable((bool) $row['gradable']);
                $refFiles[] = $refFile;
            }
        }
        $refFiles = self::sortByShortFileName($refFiles);
        return $refFiles;
    }
    public function isGradable($shortFileName)
    {
        if (empty(self::$gradables)) {
            $resources = $this->get(['filetype' => 'resources', 'gradable' => 1]);
            foreach ($resources as $resource) {
                self::$gradables[] = $resource->getShortFileName();
            }
        }
        return in_array($shortFileName, self::$gradables);
    }
    public function getByObject($subQuestion, $fileType = '')
    {
        $where = ['subquestion_name' => $subQuestion->getName()];
        if ($fileType) {
            $where['filetype'] = $fileType;
        }
        return $this->get($where);
    }
    private function put($subQuestion, $refFiles)
    {
        $table = self::$table;
        $subQuestionName = $subQuestion->getName();
        if (count($refFiles) <= 0) { // No answers found. An error!
            HT::error("<strong>SubQuestion $subQuestionName</strong>: 
            No solutions or resources found!");
        } else {
            $columns = [
                'subquestion_name',
                'filetype',
                'filename',
                'fulltext',
                'hash0',
                'hash1',
                'sim',
                'gradable'
            ];
            $rows = [];
            foreach ($refFiles as $refFile) {
                $fileName = $refFile->getFileName();
                $fileType = $refFile->getFileType();
                $fullText = DB::$db->getFileContent($fileName);
                $hash0 = $refFile->getHash0();
                $hash1 = $refFile->getHash1();
                $sim = (float) $refFile->getSim();
                $gradable = (int) $refFile->isGradable();
                $rows[] = [
                    $subQuestionName,
                    $fileType,
                    $fileName,
                    $fullText,
                    $hash0,
                    $hash1,
                    $sim,
                    $gradable
                ];
            }
            // DB::$db->put($table, $columns, $rows);
            DB::$db->update($table, $columns, $rows);
        }
    }
    public function process($subQuestion)
    {
        $subQuestionName = $subQuestion->getName();
        $folders = [
            'resources' => CFG::$resourceFolder,
            'solutions' => CFG::$solutionFolder,
        ];
        $solutions = $resources = $refFiles = $objects = [];
        $objects['resources'] = $objects['solutions'] = [];
        foreach ($folders as $fileType => $folder) {
            $$fileType = $this->findFiles($subQuestion, $folder);
            if (count($$fileType) <= 0) {
                HT::error("<strong>SubQuestion $subQuestionName</strong>:
                 No $fileType found in $folder!");
            } else {
                foreach ($$fileType as $fileName) {
                    if (is_file($fileName)) {
                        // $fileName = CFG::mkFileName($fileName);
                        $refFile = new RefFile($subQuestionName, $fileType, $fileName);
                        $refFiles[] = $refFile;
                        $objects[$fileType][] = $refFile;
                    }
                }
            }
        }
        // We now have all the resource files ($resources) and solutions ($solutions)
        // for this question. Find the ones that are different as gradable
        foreach ($objects['resources'] as $resource) {
            $resourceShort = $resource->getShortFileName();
            foreach ($objects['solutions'] as $solution) {
                $solutionShort = $solution->getShortFileName();
                if ($solutionShort == $resourceShort) {
                    if ($resource->getHash0() == $solution->getHash0()) {
                        $sim = 100;
                    } else if ($resource->getHash1() == $solution->getHash1()) {
                        $sim = 100;
                    } else {
                        $sim = Report::similarity(
                            $resource->getFileName(),
                            $solution->getFileName()
                        );
                    }
                    $gradable = $sim < 99.9;
                    $resource->setSim($sim);
                    $solution->setSim($sim);
                    $resource->setGradable($gradable);
                    $solution->setGradable($gradable);
                }
            }
        }
        $this->put($subQuestion, $refFiles);
    }
    private function findFiles($subQuestion, $folder)
    {
        # Look for files like folder/q1/* and /q1/*/* and q1/*/*/*
        $questionNum = $subQuestion->getQuestionNum();
        $pattern = $folder . "q$questionNum/{*.*,*/*.*,*/*/*.*}";
        // $found = glob(strtolower($pattern), GLOB_BRACE);
        $found = glob($pattern, GLOB_BRACE);
        $files = [];
        $allowed = ['php', 'html'];
        foreach ($found as $f) {
            $ext = pathinfo($f, PATHINFO_EXTENSION);
            if (in_array($ext, $allowed)) {
                $files[] = $f;
            }
        }
        return $files;
    }
    public function getRefFile($shortName, $fileType)
    {
        $refFiles = $this->get(['filetype' => $fileType]);
        $found = [];
        foreach ($refFiles as $refFile) {
            if (stripos($refFile->getFileName(), "/$shortName") !== false) {
                $found[] = $refFile;
            }
        }
        $count = count($found);
        if ($count > 0) {
            for ($i = 1; $i < $count; $i++) {
                $prevName = $found[$i - 1]->getFileName();
                $curName = $found[$i]->getFileName();
                if ($prevName != $curName) {
                    HT::error("Confusion with $fileType for $shortName:
                        $prevName != $curName");
                    die;
                }
            }
            return $found[0];
        }
        return false;
    }
}