<?php
class Exam extends HT
{
    private $name;
    private $questions = [];
    private $sections = [];
    protected static $class = __CLASS__;
    protected static $actions = [
        'view' => 'Exam Details',
        'process' => 'Process Everything!',
        'rubrics' => 'Questions and Rubrics',
        'refFiles' => 'Resources and Solutions',
        'submissions' => 'Submissions',
        'delete' => 'Delete Submissions'
    ];
    protected static $actionProperties = [
        'view' => [
            'force' => 'force',
            'class' => 'error'
        ],
        'rubrics' => [
            'class' => 'warn'
        ],
        'refFiles' => [
            'class' => 'warn',
            'target' => 'right'
        ],
        'submissions' => [
            'class' => 'warn'
        ],
        'delete' => [
            'class' => 'error',
            'target' => 'right'
        ]
    ];
    protected static $intro = "<h4>Process Menu</h4>
    Here, you can process the files to set up Grader:
    <ul>
    <li>Process grade book downloads and <code>rubrics.csv</code> to create 
    questions, subQuestions and rubrics.</li>
    <li>Process the reference resource files and solutions.</li>
    <li>Process student submissions, as downloaded from eLearn.</li>
    </ul>";
    protected static $actionDesc = [
        'view' => 'Examine the questions, subQuestions and their marks as 
        processed from the eLearn grade book exported (CSV) files.
        <span style="color:red;">In Finals Mode (as opposed to Lab-Test Mode),
        this button reprocesses the eLearn output file, and repopulates
        the DB. Use with care!</span>',
        'process' => 'Does what the next three buttons do:
        Import <code>rubrics.csv</code> to create rubrics, import
        the grade book downloads from eLearn to create questions, subQuestions 
        and students, and import student submissions to create answers and 
        prepare for grading. You can also use Grader to verify that students 
        have submitted valid zip files: Grader will produce color-coded status 
        messages per student.',
        'rubrics' => 'Import <code>rubrics.csv</code> to create questions, 
        subQuestions and their rubrics in the database.',
        'refFiles' => 'Process the resource and solution files into the database',
        'submissions' => 'Import student submissions to create answers and prepare
        for grading. This button can be used to validate the submitted zip files.',
        'delete' => 'Delete the student submissions: The uncompressed files, not 
        the zip files downloaded from eLearn.'
    ];
    protected static $display = [
        "class" => "info",
        "title" => "Preprocess Files"
    ];
    public function __construct()
    {
        $this->name = CFG::$examShortName;
        if (CFG::isQuiz()) {
            $this->mkFinals();
        } else {
            $this->mkLabTest();
        }
    }
    private function mkLabTest()
    {
        // Section Names
        $sectionNames = CFG::$sectionNames;
        // Locations of subQuestion number and marks in the CSV file header
        $locSubQ = CFG::$locSubQ;
        $locMarks = CFG::$locMarks;
        // From the header row of the given CSV file (eLearn grade export),
        // construct the question and its subQuestions
        // $csv = "./CSV/IS113-LT1-{$sectionNames[0]}.csv";
        $csv = CFG::getCsvFileName($sectionNames[0]);
        if (!is_file($csv)) {
            HT::error("Error loading $csv");
            return;
        }
        $file = fopen($csv, 'r');
        $line = fgetcsv($file);
        fclose($file);
        // Filter only the questions column headers
        foreach ($line as $key => $value) {
            if (empty($value) || (preg_match('@[qQ].[ A-Za-z]@', $value) == 0)) {
                unset($line[$key]);
            }
        }
        // Get the marks for each subQuestion and question numbers
        $marks = $questionNumbers = [];
        foreach ($line as $key => $value) {
            $tokens = explode(" ", $value);
            $subQuestionName = strtolower(trim($tokens[$locSubQ]));
            $marks[$subQuestionName] = explode(":", $tokens[$locMarks])[1];
            // Use preg_match() function to check match
            preg_match('!\d+!', $subQuestionName, $match);
            if (!in_array($match[0], $questionNumbers)) {
                $questionNumbers[] = $match[0];
            }
        }
        // We now have question numbers, subQuestion names and marks
        // Add questions (which will add subQuestions too)
        // Add only gradeable questions
        foreach ($questionNumbers as $qNum) {
            if (in_array($qNum, CFG::$gradeQuestions)) {
                $this->questions[$qNum] = new Question($qNum, $marks);
            }
        }
        // Add sections (which will add students too)
        foreach ($sectionNames as $sectionName) {
            $this->sections[] = new Section($sectionName);
        }
    }
    private function mkFinals()
    {
        $sectionName = "All";
        if (empty($_POST['force'])) { // Use DB
            $students = self::$studentDao->getStudents();
            $section = new Section($sectionName);
            $this->sections[] = $section->setStudents($students);
            $questions = self::$questionDao->get();
            foreach ($questions as $question) {
                $qNum = $question->getNum();
                $subQuestions = self::$subQuestionDao->get(['question_num' => $qNum]);
                $question->setSubQuestions($subQuestions);
                $this->questions[$qNum] = $question;
            }
        } else {
            $csv = CFG::getCsvFileName("");
            if (!is_file($csv)) {
                HT::error("Error loading eLearn report $csv<br>
            Please run an <strong>Attempt Details</strong> report.");
                return;
            }
            // $lockFile = CFG::getSubmissionFolder($sectionName) . "mkFinal.lock";
            // if (file_exists($lockFile)) {
            //     HT::error("Students have been processed already!<br> 
            //     If you need to reprocess, first remove the lock file:<br>
            //     <code>$lockFile</code>.");
            //     return;
            // }
            // touch($lockFile);
            $file = fopen($csv, 'r');
            $headers = fgetcsv($file);
            $answerIdx = array_search('Answer', $headers);
            $questionIdx = array_search('Q #', $headers);
            $nameIdx = array_search('FirstName', $headers);
            $emailIdx = array_search('Username', $headers);
            $marksIdx = array_search('Out Of', $headers);
            $qNum = CFG::$gradeQuestions[0];
            $qName = "q$qNum";
            $zip = new ZipArchive;
            $students = [];
            while ($line = fgetcsv($file)) {
                if ($line[$questionIdx] == $qNum) {
                    $email = $line[$emailIdx];
                    $name = $line[$nameIdx];
                    $student = new Student($email, $name, $sectionName);
                    $zipFile = CFG::getSubmissionFolder($sectionName) . $email . ".zip";
                    $student->setZipFile($zipFile);
                    $res = $zip->open($zipFile, ZipArchive::CREATE);
                    if ($res === TRUE) {
                        // Trying to restore line-breaks gobbled by eLearn CSV export
                        // TODO: Need to find the right logic rather than hardcode it
                        $answer = str_replace("   ", "\n   ", $line[$answerIdx]);
                        // https://stackoverflow.com/a/710967
                        $answer = preg_replace('/^\h*\v+/m', '', $answer);
                        $zip->addFromString("$qName/Olympics.php", $answer);
                        $zip->close();
                    }
                    $marks[$qName] = $line[$marksIdx];
                    $students[] = $student;
                }
            }
            fclose($file);
            $section = new Section($sectionName);
            $this->sections[] = $section->setStudents($students);
            self::$studentDao->put($section, $students);
            $this->questions[$qNum] = new Question($qNum, $marks);
            self::$questionDao->put($this->questions);
            $subQuestion = new SubQuestion($qName, $marks[$qName], $qNum);
            self::$subQuestionDao->put([$subQuestion]);
        }
    }

    public function render($student)
    {
        $srcDoc = "";
        foreach ($this->questions as $question) {
            $srcDoc .= $question->render($student);
        }
        return $srcDoc;
    }
    public function grade()
    {
        foreach ($this->sections as $section) {
            $section->grade();
        }
    }
    public function csv()
    {
        HT::head();
        foreach ($this->sections as $section) {
            $section->csv();
        }
        HT::foot();
    }
    public function export()
    {
        HT::head();
        foreach ($this->sections as $section) {
            $section->export();
        }
        HT::foot();
    }
    public function stats()
    {
        HT::head();
        foreach ($this->sections as $section) {
            $section->stats();
        }
        HT::foot();
    }
    // Dispatch handlers
    public static function view()
    {
        self::init();
        $str = "";
        $exam = self::$exam;
        $marks = 0;
        foreach ($exam->questions as $question) {
            $marks += $question->getMarks();
            $str .= HT::info($question->print(), true);
        }
        $str .= HT::info("<h3>Total Marks = $marks</h3>", true);
        HT::head();
        HT::warn($str);
        HT::foot();
    }
    public static function process()
    {
        self::init();
        self::rubrics();
        self::refFiles();
        self::submissions();
    }
    public static function submissions()
    {
        self::init();
        HT::head();
        $exam = self::$exam;
        foreach ($exam->sections as $section) {
            HT::warn($section->getName());
            if (CFG::isQuiz()) {
            } else {
                self::$studentDao->process($section);
            }
            foreach ($section->getStudents() as $student) {
                self::$answerDao->process($student);
            }
        }
        HT::info("<h2>Error Message Summary of Submissions</h2>" .
            self::$answerDao->getProcessSummary());
        HT::foot();
    }
    public static function rubrics()
    {
        self::init();
        HT::head();
        $exam = self::$exam;
        $qtnSrc = "<table width='60%' align='center' class='warn grader'>
                <tr><th width='30%'>Question Number</th><th width='10%'></th>
                <th width='60%'>SubQuestion</th></tr>";
        $rubSrc = "";
        foreach ($exam->questions as $question) {
            self::$questionDao->process();
            $qtnSrc .= "<tr><td>" . $question->getNum() . "&emsp;(Marks:" .
                $question->getMarks() .
                ")</td><td></td><td></td></tr>";
            foreach ($question->getSubQuestions() as $subQuestion) {
                self::$subQuestionDao->process($question);
                self::$rubricDao->process($subQuestion);
                $subQuestionName = $subQuestion->getName();
                $qtnSrc .= "<tr><td></td><td>&rarr;</td>
                <td>$subQuestionName &emsp;(Marks:" . $subQuestion->getMarks() .
                    ")</td></tr>";
                $class = "success";
                $rubSrc .= "<table width='95%' align='center' class='grader $class'>
                <tr><th>Marks</th><th>Rubric (for $subQuestionName)</th>
                <th>File Name</th></tr>";
                $rubrics = self::$rubricDao->getByObject($subQuestion);
                foreach ($rubrics as $k => $rubric) {
                    $marks = $rubric->getMarks();
                    $rubricName = $rubric->getRubricName();
                    $rubricText = $rubric->getRubricText();
                    $rubricText = htmlentities($rubricText);
                    $shortFileName = $rubric->getShortFileName();
                    $rubSrc .= "<tr><td> $marks </td> 
                    <td style='text-align:left;'> $rubricName: 
                    <code>$rubricText</code></td>
                    <td><code>$shortFileName</code></td></tr>";
                }
                $rubSrc .= "</table><br>";
            }
        }
        $qtnSrc .= "</table><br>";
        HT::info($qtnSrc . $rubSrc);
        HT::foot();
    }
    private static function mkRow($subQuestion, $refFile)
    {
        if ($refFile->isGradable()) {
            $shown = "<td class='success'>Yes</td>";
        } else {
            $shown = "<td class='error'>No</td>";
        }
        $refFileName = CFG::mkFileName($refFile->getFileName());
        $row = "<tr><td>" . $subQuestion->getName() .
            $shown .
            "</td><td><code>" . $refFileName .
            "</code></td></tr>";
        return $row;
    }
    public static function refFiles()
    {
        self::init();
        HT::head();
        $exam = self::$exam;
        $resSrc = "<table width='98%' align='center' class='grader warn'>
                <tr><th colspan='100%'>Resources</th></tr>
                <tr><th width='10%'>SubQ</th><th width='10%'>Show?</th>
                <th width='80%'>File Name</th></tr>";
        $solSrc = "<table width='98%' align='center' class='grader warn'>
                <tr><th colspan='100%'>Solutions</th></tr>
                <tr><th width='10%'>SubQ</th><th width='10%'>Show?</th>
                <th width='80%'>File Name</th></tr>";
        foreach ($exam->questions as $question) {
            foreach ($question->getSubQuestions() as $subQuestion) {
                self::$refFileDao->process($subQuestion);
                $resources = self::$refFileDao->getByObject($subQuestion, 'resources');
                $solutions = self::$refFileDao->getByObject($subQuestion, 'solutions');
                foreach ($resources as $resource) {
                    $resSrc .= self::mkRow($subQuestion, $resource);
                }
                foreach ($solutions as $solution) {
                    $solSrc .= self::mkRow($subQuestion, $solution);
                }
            }
        }
        $resSrc .= "</table><br>";
        $solSrc .= "</table><br>";
        HT::info($resSrc . $solSrc);
        HT::foot();
    }
    protected static function delete()
    {
        HT::head();
        if (!isset($_POST['confirmDelete'])) {
            $str = "Deleting all unzipped student files!<br>
              Really delete them? <table align='right'><tr>
              <td><form method='post' action='?do'>
              <input type='hidden' name='do' value='delete'>
              <input type='hidden' name='confirmDelete' value='dummy'>
              <button type='submit' class='error'>
              <strong>Proceed to delete</strong>
              </button></form></td>
              </tr></table>";
            HT::warn($str);
        } else {
            HT::error("Deleting all student submissions...");
            self::init();
            $exam = self::$exam;
            foreach ($exam->getSections() as $section) {
                $section->delete();
            }
        }
        HT::foot();
    }
    public function __call($method, $args)
    {
        HT::head();
        HT::error("Got an unknown dispatch request: " . __CLASS__ .
            "->$method!");
        $this->grade();
        HT::foot();
    }
    /**
     * @return mixed
     */
    public function getQuestions()
    {
        return $this->questions;
    }
    /**
     * @return mixed
     */
    public function getSections()
    {
        return $this->sections;
    }
}